# -*- coding: utf-8 -*-
# pylint: disable=redefined-outer-name
"""Tests for the :mod:`aiida_pseudo.groups.mixins.cutoffs` module."""
import copy

import pytest

from aiida_pseudo.groups.family import CutoffsPseudoPotentialFamily


@pytest.mark.usefixtures('clear_db')
def test_get_cutoffs_dict(get_pseudo_family, generate_cutoffs_dict):
    """Test the ``CutoffsPseudoPotentialFamily._get_cutoffs_dict`` method."""
    family = get_pseudo_family(cls=CutoffsPseudoPotentialFamily)
    assert family._get_cutoffs_dict() == {}  # pylint: disable=protected-access

    for stringency, cutoffs in generate_cutoffs_dict(family).items():
        family.set_cutoffs(cutoffs, stringency)
    assert family._get_cutoffs_dict() == generate_cutoffs_dict(family)  # pylint: disable=protected-access


@pytest.mark.usefixtures('clear_db')
def test_get_cutoffs_unit_dict(get_pseudo_family, generate_cutoffs_dict):
    """Test the ``CutoffsPseudoPotentialFamily._get_cutoffs_unit_dict`` method."""
    family = get_pseudo_family(cls=CutoffsPseudoPotentialFamily)
    assert family._get_cutoffs_unit_dict() == {}  # pylint: disable=protected-access

    default_units_dict = {}
    for stringency, cutoffs in generate_cutoffs_dict(family).items():
        family.set_cutoffs(cutoffs, stringency)
        default_units_dict[stringency] = CutoffsPseudoPotentialFamily.DEFAULT_UNIT

    assert family._get_cutoffs_unit_dict() == default_units_dict  # pylint: disable=protected-access


@pytest.mark.usefixtures('clear_db')
def test_validate_cutoffs_unit():
    """Test the ``CutoffsPseudoPotentialFamily.validate_cutoffs_unit`` method."""
    with pytest.raises(TypeError):
        CutoffsPseudoPotentialFamily.validate_cutoffs_unit(10)

    with pytest.raises(ValueError, match=r'`invalid` is not a valid unit.'):
        CutoffsPseudoPotentialFamily.validate_cutoffs_unit('invalid')

    with pytest.raises(ValueError, match=r'`watt` is not a valid energy unit.'):
        CutoffsPseudoPotentialFamily.validate_cutoffs_unit('watt')


@pytest.mark.usefixtures('clear_db')
def test_validate_stringency(get_pseudo_family, generate_cutoffs):
    """Test the ``CutoffsPseudoPotentialFamily.validate_stringency`` method."""
    family = get_pseudo_family(cls=CutoffsPseudoPotentialFamily)

    with pytest.raises(ValueError, match=r'stringency `.*` is not one of the available cutoff stringencies for this'):
        family.validate_stringency('default')

    cutoffs = generate_cutoffs(family)
    stringency = 'default'
    family.set_cutoffs(cutoffs, stringency)

    with pytest.raises(ValueError, match=r'stringency `.*` is not one of the available cutoff stringencies for this'):
        family.validate_stringency(stringency + 'non-existing')

    family.validate_stringency(stringency)


@pytest.mark.usefixtures('clear_db')
def test_get_default_stringency(get_pseudo_family, generate_cutoffs):
    """Test the ``CutoffsPseudoPotentialFamily.get_default_stringency`` method."""
    family = get_pseudo_family(cls=CutoffsPseudoPotentialFamily)

    with pytest.raises(ValueError, match='no default stringency has been defined.'):
        family.get_default_stringency()

    cutoffs = generate_cutoffs(family)
    stringency = 'default'
    family.set_cutoffs(cutoffs, stringency)

    assert family.get_default_stringency() == stringency


@pytest.mark.usefixtures('clear_db')
def test_set_default_stringency(get_pseudo_family, generate_cutoffs_dict):
    """Test the ``CutoffsPseudoPotentialFamily.set_default_stringency`` method."""
    family = get_pseudo_family(cls=CutoffsPseudoPotentialFamily)
    assert family.get_cutoff_stringencies() == ()

    stringencies = ('low', 'normal')
    for stringency, cutoffs in generate_cutoffs_dict(family, stringencies).items():
        family.set_cutoffs(cutoffs, stringency)

    assert family.get_default_stringency() == 'low'

    with pytest.raises(ValueError, match=r'the default stringency cannot be unset'):
        family.set_default_stringency(None)

    with pytest.raises(ValueError, match=r'stringency `nonexistent` is not one of the available cutoff stringencies'):
        family.set_default_stringency('nonexistent')

    family.set_default_stringency('normal')
    assert family.get_default_stringency() == 'normal'


@pytest.mark.usefixtures('clear_db')
def test_get_cutoff_stringencies(get_pseudo_family, generate_cutoffs_dict):
    """Test the ``CutoffsPseudoPotentialFamily.get_cutoff_stringencies`` method."""
    family = get_pseudo_family(cls=CutoffsPseudoPotentialFamily)
    assert family.get_cutoff_stringencies() == ()

    stringencies = ('low', 'normal', 'high')
    for stringency, cutoffs in generate_cutoffs_dict(family, stringencies).items():
        family.set_cutoffs(cutoffs, stringency)

    assert sorted(family.get_cutoff_stringencies()) == sorted(stringencies)


@pytest.mark.usefixtures('clear_db')
def test_set_cutoffs(get_pseudo_family, generate_cutoffs):
    """Test the ``CutoffsPseudoPotentialFamily.set_cutoffs`` method."""
    elements = ['Ar', 'He']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffsPseudoPotentialFamily, elements=elements)
    cutoffs = generate_cutoffs(family)
    stringency = 'default'

    family.set_cutoffs(cutoffs, stringency)
    assert family.get_cutoffs() == cutoffs
    assert family.get_cutoffs(stringency) == cutoffs

    with pytest.raises(ValueError, match=r'cutoffs defined for unsupported elements: .*'):
        cutoffs_invalid = copy.deepcopy(cutoffs)
        cutoffs_invalid['C'] = {'cutoff_wfc': 1.0, 'cutoff_rho': 2.0}
        family.set_cutoffs(cutoffs_invalid, stringency)

    with pytest.raises(ValueError, match=r'cutoffs not defined for all family elements: .*'):
        cutoffs_invalid = copy.deepcopy(cutoffs)
        cutoffs_invalid.pop('He')
        family.set_cutoffs(cutoffs_invalid, stringency)

    with pytest.raises(ValueError, match=r'invalid cutoff keys for element .*: .*'):
        cutoffs_invalid = copy.deepcopy(cutoffs)
        cutoffs_invalid['He'] = {'cutoff_wfc': 1.0}
        family.set_cutoffs(cutoffs_invalid, stringency)

    with pytest.raises(ValueError, match=r'invalid cutoff keys for element .*: .*'):
        cutoffs_invalid = copy.deepcopy(cutoffs)
        cutoffs_invalid['He'] = {'cutoff_wfc': 1.0, 'cutoff_rho': 2.0, 'cutoff_extra': 3.0}
        family.set_cutoffs(cutoffs_invalid, stringency)

    with pytest.raises(ValueError, match=r'invalid cutoff values for element .*: .*'):
        cutoffs_invalid = copy.deepcopy(cutoffs)
        cutoffs_invalid['He'] = {'cutoff_wfc': 1.0, 'cutoff_rho': 'string'}
        family.set_cutoffs(cutoffs_invalid, stringency)


@pytest.mark.usefixtures('clear_db')
def test_set_cutoffs_unit_default(get_pseudo_family, generate_cutoffs):
    """Test the ``CutoffsPseudoPotentialFamily.set_cutoffs`` sets a default unit if not specified."""
    elements = ['Ar']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffsPseudoPotentialFamily, elements=elements)
    cutoffs = generate_cutoffs(family)
    stringency = 'default'

    family.set_cutoffs(cutoffs, stringency)
    assert family.get_cutoffs_unit() == CutoffsPseudoPotentialFamily.DEFAULT_UNIT


@pytest.mark.usefixtures('clear_db')
def test_set_cutoffs_multiple_units(get_pseudo_family, generate_cutoffs):
    """Test ``CutoffsPseudoPotentialFamily.set_cutoffs`` correctly sets separate units for different stringencies."""
    elements = ['Ar']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffsPseudoPotentialFamily, elements=elements)
    cutoffs = generate_cutoffs(family)

    cutoffs_unit_dict = {}

    stringency = 'default'
    unit = 'hartree'
    family.set_cutoffs(cutoffs, stringency, unit)
    cutoffs_unit_dict[stringency] = unit
    assert family.get_cutoffs_unit() == unit

    stringency = 'rydberg'
    unit = 'Ry'
    family.set_cutoffs(cutoffs, stringency, unit)
    cutoffs_unit_dict[stringency] = unit
    assert family.get_cutoffs_unit(stringency) == unit

    # pylint: disable=protected-access
    assert family._get_cutoffs_unit_dict() == {'default': 'hartree', 'rydberg': 'Ry'}


@pytest.mark.usefixtures('clear_db')
def test_set_cutoffs_auto_default(get_pseudo_family, generate_cutoffs):
    """Test that the ``CutoffsPseudoPotentialFamily.set_cutoffs`` method specifies the correct default stringency.

    If family only has one stringency specified, the `set_cutoffs` method automatically sets this as the default.
    """
    elements = ['Ar']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffsPseudoPotentialFamily, elements=elements)
    cutoffs = generate_cutoffs(family)
    stringency = 'default'

    family.set_cutoffs(cutoffs, stringency)
    assert family.get_cutoffs() == cutoffs
    assert family.get_default_stringency() == 'default'


@pytest.mark.usefixtures('clear_db')
def test_get_cutoffs(get_pseudo_family, generate_cutoffs):
    """Test the ``CutoffsPseudoPotentialFamily.get_cutoffs`` method."""
    elements = ['Ar', 'He']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffsPseudoPotentialFamily, elements=elements)
    cutoffs = generate_cutoffs(family)
    stringency = 'default'

    with pytest.raises(ValueError, match='no default stringency has been defined.'):
        family.get_cutoffs()

    family.set_cutoffs(cutoffs, stringency)

    with pytest.raises(ValueError, match=r'stringency `.*` is not one of the available cutoff stringencies for this'):
        family.get_cutoffs('non-existing')

    assert family.get_cutoffs() == cutoffs

    low_cutoffs = {element: {'cutoff_wfc': 0.5, 'cutoff_rho': 1.0} for element in elements}
    family.set_cutoffs(low_cutoffs, 'low')

    assert family.get_cutoffs('low') == low_cutoffs


@pytest.mark.usefixtures('clear_db')
def test_get_recommended_cutoffs(get_pseudo_family, generate_structure, generate_cutoffs):
    """Test the ``CutoffsPseudoPotentialFamily.get_recommended_cutoffs`` method."""
    elements = ['Ar', 'He']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffsPseudoPotentialFamily, elements=elements)
    cutoffs = generate_cutoffs(family)
    stringency = 'default'
    family.set_cutoffs(cutoffs, stringency)

    structure = generate_structure(elements=elements)

    with pytest.raises(ValueError):
        family.get_recommended_cutoffs(elements=None, structure=None)

    with pytest.raises(ValueError):
        family.get_recommended_cutoffs(elements='Ar', structure=structure)

    with pytest.raises(TypeError):
        family.get_recommended_cutoffs(elements=False, structure=None)

    with pytest.raises(TypeError):
        family.get_recommended_cutoffs(elements=None, structure='Ar')

    expected = cutoffs['Ar']
    assert family.get_recommended_cutoffs(elements='Ar') == (expected['cutoff_wfc'], expected['cutoff_rho'])
    assert family.get_recommended_cutoffs(elements=('Ar',)) == (expected['cutoff_wfc'], expected['cutoff_rho'])

    expected = cutoffs['He']
    assert family.get_recommended_cutoffs(elements=('Ar', 'He')) == (expected['cutoff_wfc'], expected['cutoff_rho'])

    expected = cutoffs['He']
    assert family.get_recommended_cutoffs(structure=structure) == (expected['cutoff_wfc'], expected['cutoff_rho'])

    # Try structure with multiple kinds with the same element
    expected = cutoffs['He']
    structure = generate_structure(elements=['He1', 'He2'])
    assert family.get_recommended_cutoffs(structure=structure) == (expected['cutoff_wfc'], expected['cutoff_rho'])


@pytest.mark.usefixtures('clear_db')
def test_get_recommended_cutoffs_unit(get_pseudo_family, generate_cutoffs):
    """Test the ``CutoffsPseudoPotentialFamily.get_recommended_cutoffs`` method with the ``unit`` argument."""
    elements = ['Ar', 'He']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffsPseudoPotentialFamily, elements=elements)
    cutoffs = generate_cutoffs(family)
    stringency = 'default'
    unit = 'Eh'
    family.set_cutoffs(cutoffs, stringency, unit)

    cutoffs_ar = cutoffs['Ar']

    expected = (cutoffs_ar['cutoff_wfc'], cutoffs_ar['cutoff_rho'])
    assert family.get_recommended_cutoffs(elements='Ar') == expected

    expected = (cutoffs_ar['cutoff_wfc'] * 2, cutoffs_ar['cutoff_rho'] * 2)
    assert family.get_recommended_cutoffs(elements='Ar', unit='Ry') == expected


@pytest.mark.usefixtures('clear_db')
def test_get_cutoffs_unit(get_pseudo_family, generate_cutoffs):
    """Test the ``CutoffsPseudoPotentialFamily.get_cutoffs_unit`` method."""
    elements = ['Ar', 'He']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffsPseudoPotentialFamily, elements=elements)
    cutoffs = generate_cutoffs(family)
    stringency = 'default'
    family.set_cutoffs(cutoffs, stringency)

    assert family.get_cutoffs_unit() == 'eV'
    cutoffs = family.get_cutoffs()
    stringency = family.get_default_stringency()

    family.set_cutoffs(cutoffs, stringency, unit='Ry')
    assert family.get_cutoffs_unit() == 'Ry'

    family.set_cutoffs(cutoffs, stringency, unit='Eh')
    assert family.get_cutoffs_unit() == 'Eh'


@pytest.mark.usefixtures('clear_db')
def test_delete_cutoffs(get_pseudo_family, generate_cutoffs_dict):
    """Test the ``CutoffsPseudoPotentialFamily.delete_cutoffs`` method."""
    elements = ['Ar', 'He']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffsPseudoPotentialFamily, elements=elements)

    stringencies = ('low', 'normal', 'high')
    for stringency, cutoffs in generate_cutoffs_dict(family, stringencies).items():
        family.set_cutoffs(cutoffs, stringency)

    with pytest.raises(ValueError, match=r'stringency `nonexistent` is not one of the available cutoff stringencies'):
        family.delete_cutoffs('nonexistent')

    with pytest.warns(UserWarning, match='`low` was the default stringency of this family. Please set'):
        family.delete_cutoffs('low')
    assert sorted(family.get_cutoff_stringencies()) == sorted(('normal', 'high'))

    with pytest.raises(ValueError, match='no default stringency has been defined.'):
        family.get_default_stringency()

    with pytest.warns(UserWarning, match='Setting `high` as the default since it is now the only defined stringency.'):
        family.delete_cutoffs('normal')

    assert family.get_default_stringency() == 'high'



# -*- coding: utf-8 -*-
# pylint: disable=redefined-outer-name
"""Tests that are common to all data plugins in the :py:mod:`~aiida_pseudo.data.pseudo` module."""
import pytest

from aiida import plugins


def get_entry_point_names():
    """Return the registered entry point names for the given common workflow.

    :param workflow: the name of the common workflow.
    :param leaf: if True, only return the leaf of the entry point name, i.e., the name of plugin that implements it.
    :return: list of entry points names.
    """
    prefix = 'pseudo.'
    entry_points_names = plugins.entry_point.get_entry_point_names('aiida.data')
    return [name for name in entry_points_names if name.startswith(prefix)]


@pytest.fixture(scope='function', params=get_entry_point_names())
def entry_point_name(request):
    """Fixture that parametrizes over all the registered subclass implementations of ``PseudoPotentialData``."""
    return request.param


def test_get_entry_point_name(entry_point_name):
    """Test the ``PseudoPotentialData.get_entry_point_name`` method."""
    cls = plugins.DataFactory(entry_point_name)
    assert cls.get_entry_point_name() == entry_point_name



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument,redefined-outer-name
"""Tests for the command `aiida-pseudo family`."""
import json

from numpy.testing import assert_almost_equal
import pytest

from aiida.orm import Group

from aiida_pseudo.cli.family import cmd_family_cutoffs_set, cmd_family_show
from aiida_pseudo.groups.family import PseudoPotentialFamily, CutoffsPseudoPotentialFamily


@pytest.mark.usefixtures('clear_db')
def test_family_cutoffs_set(run_cli_command, get_pseudo_family, generate_cutoffs_dict, tmp_path):
    """Test the `aiida-pseudo family cutoffs set` command."""
    family = get_pseudo_family(cls=CutoffsPseudoPotentialFamily)
    stringencies = ('low', 'normal', 'high')
    cutoffs_dict = generate_cutoffs_dict(family, stringencies=stringencies)

    # Set two stringencies
    for stringency in ('low', 'normal'):
        family.set_cutoffs(cutoffs_dict[stringency], stringency)
    assert sorted(family.get_cutoff_stringencies()) == sorted(['low', 'normal'])

    filepath = tmp_path / 'cutoffs.json'

    # Invalid JSON
    filepath.write_text('invalid content')
    result = run_cli_command(cmd_family_cutoffs_set, [family.label, str(filepath)], raises=True)
    assert "Error: Missing option '-s' / '--stringency'" in result.output
    assert sorted(family.get_cutoff_stringencies()) == sorted(['low', 'normal'])

    # Invalid cutoffs structure
    filepath.write_text(json.dumps({'Ar': {'cutoff_rho': 300}}))
    result = run_cli_command(cmd_family_cutoffs_set, [family.label, str(filepath), '-s', 'high'], raises=True)
    assert 'Error: Invalid value for CUTOFFS:' in result.output
    assert sorted(family.get_cutoff_stringencies()) == sorted(['low', 'normal'])

    # Set correct stringency
    stringency = 'high'
    filepath.write_text(json.dumps(cutoffs_dict['high']))
    result = run_cli_command(cmd_family_cutoffs_set, [family.label, str(filepath), '-s', stringency])
    assert 'Success: set cutoffs for' in result.output
    assert stringency in family.get_cutoff_stringencies()
    assert sorted(family.get_cutoff_stringencies()) == sorted(['low', 'normal', 'high'])
    assert family.get_cutoffs(stringency) == cutoffs_dict[stringency]


@pytest.mark.usefixtures('clear_db')
def test_family_cutoffs_set_unit(run_cli_command, get_pseudo_family, generate_cutoffs, tmp_path):
    """Test the `aiida-pseudo family cutoffs set` command with the ``--unit`` flag."""
    family = get_pseudo_family(cls=CutoffsPseudoPotentialFamily)
    stringency = 'normal'
    cutoffs = generate_cutoffs(family)

    # We explicitly set the unit to the default (eV) in case this is changed in the future to hartree.
    family.set_cutoffs(cutoffs, stringency, 'eV')

    filepath = tmp_path / 'cutoffs.json'
    filepath.write_text(json.dumps(cutoffs))

    # Invalid unit
    unit = 'GME stock'
    result = run_cli_command(
        cmd_family_cutoffs_set, [family.label, str(filepath), '-s', stringency, '-u', unit], raises=True
    )
    assert "Error: Invalid value for '-u' / '--unit': `GME stock` is not a valid unit." in result.output

    # Non-energy unit
    unit = 'second'
    result = run_cli_command(
        cmd_family_cutoffs_set, [family.label, str(filepath), '-s', stringency, '-u', unit], raises=True
    )
    assert "Error: Invalid value for '-u' / '--unit': `second` is not a valid `energy` unit." in result.output

    # Correct unit
    unit = 'hartree'
    result = run_cli_command(cmd_family_cutoffs_set, [family.label, str(filepath), '-s', stringency, '-u', unit])
    assert family.get_cutoffs_unit() == unit


def test_family_show(clear_db, run_cli_command, get_pseudo_family):
    """Test the `aiida-pseudo show` command."""
    family = get_pseudo_family()
    result = run_cli_command(cmd_family_show, [family.label])
    for node in family.nodes:
        assert node.md5 in result.output
        assert node.element in result.output
        assert node.filename in result.output


def test_family_show_recommended_cutoffs(clear_db, run_cli_command, get_pseudo_family, generate_cutoffs_dict):
    """Test the `aiida-pseudo show` command for a family with recommended cutoffs."""
    family = get_pseudo_family(cls=CutoffsPseudoPotentialFamily)
    stringencies = ('normal', 'high')
    cutoffs_dict = generate_cutoffs_dict(family, stringencies=stringencies)

    for stringency, cutoffs in cutoffs_dict.items():
        family.set_cutoffs(cutoffs, stringency)

    # Test the command prints an error for a non-existing stringency
    result = run_cli_command(cmd_family_show, [family.label, '--stringency', 'non-existing'], raises=True)
    assert "Error: Invalid value for '-s' / '--stringency': stringency `non-existing` is not" in result.output

    # Test the command for default and explicit stringency
    for stringency in [None, 'high']:

        if stringency is not None:
            result = run_cli_command(cmd_family_show, [family.label, '--stringency', stringency, '--raw'])
        else:
            result = run_cli_command(cmd_family_show, [family.label, '--raw'])

        for index, node in enumerate(sorted(family.nodes, key=lambda pseudo: pseudo.element)):
            fields = result.output_lines[index].split()
            cutoffs = family.get_recommended_cutoffs(elements=node.element, stringency=stringency)
            assert node.element in fields[0]
            assert node.filename in fields[1]
            assert node.md5 in fields[2]
            assert_almost_equal(cutoffs[0], float(fields[3]))
            assert_almost_equal(cutoffs[1], float(fields[4]))


def test_family_show_argument_type(clear_db, run_cli_command, get_pseudo_family):
    """Test that `aiida-pseudo show` only accepts instances of `PseudoPotentialFamily` or subclasses as argument."""
    pseudo_family = get_pseudo_family(label='pseudo-family', cls=PseudoPotentialFamily)
    normal_group = Group('normal-group').store()

    run_cli_command(cmd_family_show, [pseudo_family.label])
    run_cli_command(cmd_family_show, [normal_group.label], raises=SystemExit)


def test_family_show_raw(clear_db, run_cli_command, get_pseudo_family):
    """Test the `-r/--raw` option."""
    family = get_pseudo_family()

    for option in ['-r', '--raw']:
        result = run_cli_command(cmd_family_show, [option, family.label])
        assert len(result.output_lines) == len(family.nodes)


def test_family_show_unit_default(clear_db, run_cli_command, get_pseudo_family):
    """Test the `family show` command with default unit."""
    elements = ['Ar', 'Kr']
    cutoff_dict = {'normal': {'Ar': {'cutoff_wfc': 50, 'cutoff_rho': 200}, 'Kr': {'cutoff_wfc': 25, 'cutoff_rho': 100}}}

    family = get_pseudo_family(cls=CutoffsPseudoPotentialFamily, elements=elements, cutoffs_dict=cutoff_dict, unit='Ry')

    # Test the default unit (Ry)
    result = run_cli_command(cmd_family_show, [family.label])

    header_fields = result.output_lines[0].split()
    unit = family.get_cutoffs_unit()

    assert header_fields[4] == f'({unit})'
    assert header_fields[7] == f'({unit})'

    for index, element in enumerate(elements):
        cutoffs = family.get_recommended_cutoffs(elements=element)
        fields = result.output_lines[index + 2].split()
        assert_almost_equal(cutoffs[0], float(fields[3]))
        assert_almost_equal(cutoffs[1], float(fields[4]))


@pytest.mark.parametrize('unit', ['Ry', 'eV', 'hartree', 'aJ'])
def test_family_show_unit(clear_db, run_cli_command, get_pseudo_family, unit):
    """Test the `-u/--unit` option."""
    elements = [
        'Ar',
    ]
    cutoff_dict = {'normal': {'Ar': {'cutoff_wfc': 50, 'cutoff_rho': 200}}}

    family = get_pseudo_family(cls=CutoffsPseudoPotentialFamily, elements=elements, cutoffs_dict=cutoff_dict, unit='Ry')

    # Test both option strings and several units
    for option in ['-u', '--unit']:
        result = run_cli_command(cmd_family_show, [family.label, option, unit])
        cutoffs = family.get_recommended_cutoffs(elements='Ar', unit=unit)
        header_fields = result.output_lines[0].split()
        assert header_fields[4] == f'({unit})'
        assert header_fields[4] == f'({unit})'

        values_fields = result.output_lines[2].split()
        assert round(cutoffs[0], 1) == float(values_fields[3])
        assert round(cutoffs[1], 1) == float(values_fields[4])



# -*- coding: utf-8 -*-
# pylint: disable=redefined-outer-name
"""Tests for the :py:mod:`~aiida_pseudo.data.pseudo.pseudo` module."""
import io
import pathlib

import pytest

from aiida.common.files import md5_from_filelike
from aiida.common.exceptions import ModificationNotAllowed, StoringNotAllowed
from aiida.common.links import LinkType
from aiida.orm import CalcJobNode

from aiida_pseudo.data.pseudo import PseudoPotentialData, UpfData


@pytest.fixture
def source(request, filepath_pseudos):
    """Return a pseudopotential, eiter as ``str``, ``Path`` or ``io.BytesIO``."""
    filepath_pseudo = pathlib.Path(filepath_pseudos()) / 'Ar.upf'

    if request.param is str:
        return str(filepath_pseudo)

    if request.param is pathlib.Path:
        return filepath_pseudo

    return io.BytesIO(filepath_pseudo.read_bytes())


@pytest.mark.parametrize('source', (io.BytesIO, str, pathlib.Path), indirect=True)
def test_constructor_source_types(source):
    """Test the constructor."""
    pseudo = PseudoPotentialData(source)
    assert isinstance(pseudo, PseudoPotentialData)
    assert not pseudo.is_stored


def test_constructor_invalid():
    """Test the constructor for invalid arguments."""
    with pytest.raises(TypeError, match='missing 1 required positional argument'):
        PseudoPotentialData()  # pylint: disable=no-value-for-parameter


@pytest.mark.usefixtures('chtmpdir')
@pytest.mark.parametrize('source_type', ('stream', 'str_absolute', 'str_relative', 'pathlib.Path'))
@pytest.mark.parametrize('implicit', (True, False))
def test_constructor_filename(get_pseudo_potential_data, implicit, source_type):
    """Test the ``filename`` argument of the constructor."""
    pseudo = get_pseudo_potential_data()
    explicit_filename = 'custom.dat'

    # Copy the content of the test pseudo to file in the current working directory
    filepath = pathlib.Path('tempfile.pseudo')

    with open(filepath, mode='wb') as handle:
        handle.write(pseudo.get_object_content(pseudo.filename, mode='rb'))
        handle.flush()

    if source_type == 'stream':
        with open(filepath, 'rb') as handle:
            source = io.BytesIO(handle.read())
    elif source_type == 'str_absolute':
        source = str(filepath.absolute())
    elif source_type == 'str_relative':
        source = str(filepath.name)
    elif source_type == 'pathlib.Path':
        source = filepath

    if implicit:
        node = PseudoPotentialData(source, filename=None)
        # If the source type was a stream, we pass a bytestream which doesn't have a name and so the name will be
        # determined by the baseclass which has some default, but in this case we don't have to check anything.
        if source_type != 'stream':
            assert node.filename == filepath.name
    else:
        node = PseudoPotentialData(source, filename=explicit_filename)
        assert node.filename == explicit_filename


@pytest.mark.parametrize(('value, exception, pattern'), (
    (io.StringIO('content'), TypeError, r'`source` should be a `str` or `pathlib.Path` filepath .*'),
    ('non-existing/path', FileNotFoundError, r'No such file or directory: .*'),
    (pathlib.Path('non-existing/path'), FileNotFoundError, r'No such file or directory: .*'),
))
def test_prepare_source_excepts(value, exception, pattern):
    """Test the ``PseudoPotentialData.prepare_source`` method when it is supposed to except."""
    with pytest.raises(exception, match=pattern):
        PseudoPotentialData.prepare_source(value)


@pytest.mark.parametrize('source', (io.BytesIO, str, pathlib.Path), indirect=True)
def test_prepare_source(source):
    """Test the ``PseudoPotentialData.prepare_source`` method for valid input."""
    assert isinstance(PseudoPotentialData.prepare_source(source), io.BytesIO)

    if isinstance(source, io.BytesIO):
        # If we pass a bytestream, we should get the exact same back
        assert PseudoPotentialData.prepare_source(source) is source


@pytest.mark.usefixtures('clear_db')
def test_store():
    """Test the `PseudoPotentialData.store` method."""
    stream = io.BytesIO(b'pseudo')
    md5_correct = md5_from_filelike(stream)
    md5_incorrect = 'abcdef0123456789'
    stream.seek(0)

    pseudo = PseudoPotentialData(io.BytesIO(b'pseudo'))

    with pytest.raises(StoringNotAllowed, match='no valid element has been defined.'):
        pseudo.store()

    pseudo.element = 'Ar'
    pseudo.set_attribute(PseudoPotentialData._key_md5, md5_incorrect)  # pylint: disable=protected-access

    with pytest.raises(StoringNotAllowed, match=r'md5 does not match that of stored file:'):
        pseudo.store()

    pseudo.md5 = md5_correct
    result = pseudo.store()
    assert result is pseudo
    assert pseudo.is_stored


@pytest.mark.usefixtures('clear_db')
def test_element():
    """Test the `PseudoPotentialData.element` property."""
    element = 'Ar'
    pseudo = PseudoPotentialData(io.BytesIO(b'pseudo'))
    assert pseudo.element is None

    element = 'He'
    pseudo.element = element
    assert pseudo.element == element

    with pytest.raises(ValueError, match=r'.* is not a valid element'):
        pseudo.element = 'Aa'

    pseudo.store()

    with pytest.raises(ModificationNotAllowed, match='the attributes of a stored entity are immutable'):
        pseudo.element = element


@pytest.mark.usefixtures('clear_db')
def test_md5():
    """Test the `PseudoPotentialData.md5` property."""
    stream = io.BytesIO(b'pseudo')
    md5 = md5_from_filelike(stream)
    stream.seek(0)

    pseudo = PseudoPotentialData(stream)
    pseudo.element = 'Ar'
    assert pseudo.md5 == md5

    with pytest.raises(ValueError, match=r'md5 does not match that of stored file.*'):
        pseudo.md5 = 'abcdef0123456789'

    pseudo.store()

    with pytest.raises(ModificationNotAllowed, match='the attributes of a stored entity are immutable'):
        pseudo.md5 = md5


@pytest.mark.usefixtures('clear_db')
def test_store_indirect():
    """Test the `PseudoPotentialData.store` method when called indirectly because its is an input."""
    pseudo = PseudoPotentialData(io.BytesIO(b'pseudo'))
    pseudo.element = 'Ar'

    node = CalcJobNode()
    node.add_incoming(pseudo, link_type=LinkType.INPUT_CALC, link_label='pseudo')
    node.store_all()


@pytest.mark.usefixtures('clear_db')
def test_get_or_create(get_pseudo_potential_data):
    """Test the ``PseudoPotentialData.get_or_create`` classmethod."""
    upf = get_pseudo_potential_data(entry_point='upf')
    stream = io.BytesIO(upf.get_content().encode('utf-8'))

    original = PseudoPotentialData.get_or_create(stream)
    original.element = upf.element
    assert isinstance(original, PseudoPotentialData)
    assert not original.is_stored

    # Need to store it so it can actually be loaded from it by the ``get_or_create`` method
    original.store()

    # Return the stream to initial state and call again, which should return the same node.
    stream.seek(0)
    duplicate = PseudoPotentialData.get_or_create(stream)
    assert isinstance(duplicate, PseudoPotentialData)
    assert duplicate.is_stored
    assert duplicate.uuid == original.uuid

    # If the content is different, we should get a different node.
    stream.seek(0)
    different_content = PseudoPotentialData.get_or_create(io.BytesIO(b'different'))
    different_content.element = upf.element
    assert isinstance(different_content, PseudoPotentialData)
    assert not different_content.is_stored
    assert different_content.uuid != original.uuid

    # If the class is different, even if it is a subclass, we should get a different node even if content is identical
    stream.seek(0)
    different_class = UpfData.get_or_create(stream)
    assert isinstance(different_class, PseudoPotentialData)
    assert not different_class.is_stored
    assert different_class.uuid != original.uuid



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""Tests for the command `aiida-pseudo list`."""
from aiida_pseudo.cli import cmd_list
from aiida_pseudo.cli.list import PROJECTIONS_VALID
from aiida_pseudo.data.pseudo import UpfData
from aiida_pseudo.groups.family import PseudoPotentialFamily, SsspFamily


def test_list(clear_db, run_cli_command, get_pseudo_family):
    """Test the `aiida-pseudo list` command."""
    result = run_cli_command(cmd_list)
    assert 'no pseudo potential families have been installed yet: use `aiida-pseudo install`.' in result.output

    label = 'SSSP/1.0/PBE/efficiency'
    family = get_pseudo_family(label=label)
    result = run_cli_command(cmd_list)

    assert family.label in result.output


def test_list_raw(clear_db, run_cli_command, get_pseudo_family):
    """Test the `-r/--raw` option."""
    get_pseudo_family()

    for option in ['-r', '--raw']:
        result = run_cli_command(cmd_list, [option])
        assert len(result.output_lines) == 1


def test_list_project(clear_db, run_cli_command, get_pseudo_family):
    """Test the `-p/--project` option."""
    family = get_pseudo_family()

    # Test that all `PROJECTIONS_VALID` can actually be projected and don't except
    for option in ['-P', '--project']:
        run_cli_command(cmd_list, [option] + list(PROJECTIONS_VALID))

    result = run_cli_command(cmd_list, ['--raw', '-P', 'label'])
    assert len(result.output_lines) == 1
    assert family.label in result.output


def test_list_filter(clear_db, run_cli_command, get_pseudo_family):
    """Test the filtering option `-F`."""
    family_base = get_pseudo_family(label='Pseudo potential family', cls=PseudoPotentialFamily)
    family_sssp = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=SsspFamily, pseudo_type=UpfData)

    assert PseudoPotentialFamily.objects.count() == 2

    result = run_cli_command(cmd_list, ['--raw'])
    assert len(result.output_lines) == 2
    assert family_base.label in result.output
    assert family_sssp.label in result.output

    result = run_cli_command(cmd_list, ['--raw', '-F', 'pseudo.family.sssp'])
    assert len(result.output_lines) == 1
    assert family_base.label not in result.output
    assert family_sssp.label in result.output


def test_list_filter_no_result(clear_db, run_cli_command, get_pseudo_family):
    """Test the filtering option `-F` for a type for which no families exist."""
    get_pseudo_family(label='Pseudo potential family', cls=PseudoPotentialFamily)

    result = run_cli_command(cmd_list, ['--raw', '-F', 'pseudo.family.sssp'])
    assert 'no pseudo potential families found that match the filtering criteria.' in result.output



# -*- coding: utf-8 -*-
# pylint: disable=redefined-outer-name
"""Tests for the `PseudoPotentialFamily` class."""
import distutils.dir_util
import os

import pytest

from aiida.common import exceptions
from aiida.orm import QueryBuilder

from aiida_pseudo.data.pseudo import PseudoPotentialData
from aiida_pseudo.groups.family.pseudo import PseudoPotentialFamily


def test_type_string():
    """Verify the `_type_string` class attribute is correctly set to the corresponding entry point name."""
    assert PseudoPotentialFamily._type_string == 'pseudo.family'  # pylint: disable=protected-access


def test_pseudo_types():
    """Test the `PseudoPotentialFamily.pseudo_types` method."""
    assert PseudoPotentialFamily.pseudo_types == (PseudoPotentialData,)


@pytest.mark.filterwarnings('ignore:no registered entry point for')
@pytest.mark.parametrize('pseudo_types', (None, (), int))
def test_pseudo_types_validation(pseudo_types):
    """Test constructor raises if ``_pseudo_types`` is not a tuple with subclasses of ``PseudoPotentialData``."""

    class CustomFamily(PseudoPotentialFamily):
        """Test subclass that intentionally defines incorrect type for ``_pseudo_types``."""

        _pseudo_types = pseudo_types

    with pytest.raises(RuntimeError, match=r'.* should be a tuple of `PseudoPotentialData` subclasses.'):
        CustomFamily(label='custom')


@pytest.mark.usefixtures('clear_db')
def test_pseudo_type(get_pseudo_potential_data):
    """Test ``PseudoPotentialFamily.pseudo_type`` property."""
    family = PseudoPotentialFamily(label='label').store()
    assert family.pseudo_type is None

    pseudo = get_pseudo_potential_data('Ar').store()
    family.add_nodes((pseudo,))
    assert family.pseudo_type == pseudo.get_entry_point_name()

    family.clear()
    assert family.pseudo_type is None

    family.add_nodes((pseudo,))
    assert family.pseudo_type == pseudo.get_entry_point_name()

    family.remove_nodes(pseudo)
    assert family.pseudo_type is None


@pytest.mark.usefixtures('clear_db')
def test_construct():
    """Test the construction of `PseudoPotentialFamily` works."""
    label = 'label'
    family = PseudoPotentialFamily(label=label)
    assert isinstance(family, PseudoPotentialFamily)
    assert not family.is_stored
    assert family.label == label

    label = 'family'
    description = 'description'
    family = PseudoPotentialFamily(label=label, description=description)
    assert isinstance(family, PseudoPotentialFamily)
    assert not family.is_stored
    assert family.label == label
    assert family.description == description


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder(filepath_pseudos):
    """Test the `PseudoPotentialFamily.create_from_folder` class method."""
    label = 'label'
    family = PseudoPotentialFamily.create_from_folder(filepath_pseudos(), label)
    assert isinstance(family, PseudoPotentialFamily)
    assert family.is_stored
    assert family.label == label
    assert len(family.nodes) == len(os.listdir(filepath_pseudos()))


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_nested(filepath_pseudos, tmpdir):
    """Test the `PseudoPotentialFamily.create_from_folder` class method when the pseudos are in a subfolder."""
    filepath = str(tmpdir / 'subdirectory')
    distutils.dir_util.copy_tree(filepath_pseudos(), filepath)

    label = 'label'
    family = PseudoPotentialFamily.create_from_folder(str(tmpdir), label)
    assert isinstance(family, PseudoPotentialFamily)
    assert family.is_stored
    assert family.label == label
    assert len(family.nodes) == len(os.listdir(filepath_pseudos()))


@pytest.mark.usefixtures('clear_db')
@pytest.mark.parametrize('deduplicate', (True, False))
def test_create_from_folder_deduplicate(filepath_pseudos, deduplicate):
    """Test the `PseudoPotentialFamily.create_from_folder` class method."""
    from aiida_pseudo.data.pseudo.upf import UpfData
    from aiida_pseudo.groups.family.sssp import SsspFamily

    # We create an existing `PseudoPotentialFamily` as well as a `SsspFamily` to test that the deduplication mechanism
    # will only ever check for pseudo potentials of the exact same type and not allow subclasses
    original = PseudoPotentialFamily.create_from_folder(filepath_pseudos(), 'original_family')
    SsspFamily.create_from_folder(filepath_pseudos(), 'SSSP/1.0/PBE/efficiency', pseudo_type=UpfData)

    family = PseudoPotentialFamily.create_from_folder(filepath_pseudos(), 'duplicate_family', deduplicate=deduplicate)

    assert isinstance(family, PseudoPotentialFamily)
    assert family.is_stored

    pseudo_count = len(os.listdir(filepath_pseudos()))
    original_pseudos = {pseudo.pk for pseudo in original.pseudos.values()}
    family_pseudos = {pseudo.pk for pseudo in family.pseudos.values()}

    if deduplicate:
        assert QueryBuilder().append(PseudoPotentialFamily.pseudo_types, subclassing=False).count() == pseudo_count
        assert not original_pseudos.difference(family_pseudos)
    else:
        assert QueryBuilder().append(PseudoPotentialFamily.pseudo_types, subclassing=False).count() == pseudo_count * 2
        assert not original_pseudos.intersection(family_pseudos)


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_parse_fail(tmpdir):
    """Test the `PseudoPotentialFamily.create_from_folder` class method for file that fails to parse.

    Since the base pseudo potential class cannot really fail to parse, since there is no parsing, this would be
    difficult to test, however, the constructor parses the filename for the element, and that can fail if the filename
    has the incorrect format.
    """
    with open(os.path.join(str(tmpdir), 'Arr.upf'), 'wb'):
        pass

    with pytest.raises(exceptions.ParsingError, match=r'`.*` constructor did not define the element .*'):
        PseudoPotentialFamily.create_from_folder(str(tmpdir), 'label')


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_empty(tmpdir):
    """Test the `PseudoPotentialFamily.create_from_folder` class method for empty folder."""
    with pytest.raises(ValueError, match=r'no pseudo potentials were parsed from.*'):
        PseudoPotentialFamily.create_from_folder(str(tmpdir), 'label')


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_duplicate_element(tmpdir, filepath_pseudos):
    """Test the `PseudoPotentialFamily.create_from_folder` class method for folder containing duplicate element."""
    distutils.dir_util.copy_tree(filepath_pseudos(), str(tmpdir))

    with open(os.path.join(str(tmpdir), 'Ar.UPF'), 'wb'):
        pass

    with pytest.raises(ValueError, match=r'directory `.*` contains pseudo potentials with duplicate elements'):
        PseudoPotentialFamily.create_from_folder(str(tmpdir), 'label')


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_duplicate(filepath_pseudos):
    """Test that `PseudoPotentialFamily.create_from_folder` raises for duplicate label."""
    label = 'label'
    PseudoPotentialFamily(label=label).store()

    with pytest.raises(ValueError, match=r'the PseudoPotentialFamily `.*` already exists'):
        PseudoPotentialFamily.create_from_folder(filepath_pseudos(), label)


def test_parse_pseudos_from_directory_non_file(tmpdir):
    """Test the `PseudoPotentialFamily.parse_pseudos_from_directory` class method for folder containing a non-file.

    Note that a subdirectory containing the pseudos is fine, but if we find a directory and any other object at the
    base path, it should raise.
    """
    os.makedirs(os.path.join(str(tmpdir), 'directory'))
    with open(os.path.join(str(tmpdir), 'Ar.upf'), 'wb'):
        pass

    with pytest.raises(ValueError, match=r'dirpath `.*` contains at least one entry that is not a file'):
        PseudoPotentialFamily.parse_pseudos_from_directory(str(tmpdir))


def test_parse_pseudos_from_directory_non_file_nested(tmpdir):
    """Test the `PseudoPotentialFamily.parse_pseudos_from_directory` class method for folder containing a non-file.

    Note that a subdirectory containing the pseudos is fine, but if we find a directory and any other object at the
    base path, it should raise.
    """
    os.makedirs(os.path.join(str(tmpdir), 'pseudos', 'directory'))
    with open(os.path.join(str(tmpdir), 'pseudos', 'Ar.upf'), 'wb'):
        pass

    with pytest.raises(ValueError, match=r'dirpath `.*` contains at least one entry that is not a file'):
        PseudoPotentialFamily.parse_pseudos_from_directory(str(tmpdir))


@pytest.mark.filterwarnings('ignore:no registered entry point for `SomeFamily` so its instances will not be storable.')
def test_parse_pseudos_from_directory_incorrect_pseudo_type(tmpdir):
    """Test the `PseudoPotentialFamily.parse_pseudos_from_directory` for invalid ``pseudo_type`` arguments.

    It should be in ``cls._pseudo_types`` and if not explicitly defined, ``cls._pseudo_types`` should only contain a
    single element.
    """
    from aiida_pseudo.data.pseudo import PsfData, PsmlData, UpfData

    class SomeFamily(PseudoPotentialFamily):
        """Dummy pseudo family class that defines two supported pseudopotential types."""

        _pseudo_types = (PsfData, UpfData)

    with pytest.raises(ValueError, match=r'.* supports more than one type, so `pseudo_type` needs to be explicitly .*'):
        SomeFamily.parse_pseudos_from_directory(str(tmpdir))

    with pytest.raises(ValueError, match=r'`.*` is not supported by `.*`'):
        SomeFamily.parse_pseudos_from_directory(str(tmpdir), pseudo_type=PsmlData)


@pytest.mark.usefixtures('clear_db')
def test_add_nodes(get_pseudo_family, get_pseudo_potential_data):
    """Test that `PseudoPotentialFamily.add_nodes` method."""
    family = get_pseudo_family(elements=('Rn',))
    assert family.count() == 1

    pseudos = get_pseudo_potential_data('Ar').store()
    family.add_nodes(pseudos)
    assert family.count() == 2

    pseudos = (get_pseudo_potential_data('Ne').store(),)
    family.add_nodes(pseudos)
    assert family.count() == 3

    pseudos = (get_pseudo_potential_data('He').store(), get_pseudo_potential_data('Kr').store())
    family.add_nodes(pseudos)
    assert family.count() == 5

    # Test for an unstored family
    family = PseudoPotentialFamily(label='label')
    with pytest.raises(exceptions.ModificationNotAllowed):
        family.add_nodes(pseudos)


@pytest.fixture
def nodes_unstored(get_pseudo_potential_data, request):
    """Dynamic fixture returning instances of `PseudoPotentialData` either isolated or as a list."""
    if request.param == 'single':
        return get_pseudo_potential_data()

    if request.param == 'tuple':
        return (get_pseudo_potential_data(),)

    return [get_pseudo_potential_data(), get_pseudo_potential_data('Ne')]


@pytest.mark.usefixtures('clear_db')
@pytest.mark.parametrize('nodes_unstored', ['single', 'tuple', 'list'], indirect=True)
def test_add_nodes_unstored(get_pseudo_family, nodes_unstored):
    """Test that `PseudoPotentialFamily.add_nodes` fails if one or more nodes are unstored."""
    family = get_pseudo_family(elements=('He',))
    count = family.count()

    with pytest.raises(ValueError, match='At least one of the provided nodes is unstored, stopping...'):
        family.add_nodes(nodes_unstored)

    assert family.count() == count


@pytest.fixture
def nodes_incorrect_type(get_pseudo_potential_data, request):
    """Dynamic fixture returning instances of `UpfData` either isolated or as a list."""
    from aiida.orm import Data

    if request.param == 'single':
        return Data().store()

    if request.param == 'tuple':
        return (Data().store(),)

    return [get_pseudo_potential_data().store(), Data().store()]


@pytest.mark.usefixtures('clear_db')
@pytest.mark.parametrize('nodes_incorrect_type', ['single', 'tuple', 'list'], indirect=True)
def test_add_nodes_incorrect_type(get_pseudo_family, nodes_incorrect_type):
    """Test that `PseudoPotentialFamily.add_nodes` fails if one or more nodes has the incorrect type.

    Even though `UpfData` is a subclass of `PseudoPotentialData` it should still be refused because `add_nodes` checks
    for exact equality of the expected type and does not accept subclasses.
    """
    family = get_pseudo_family()
    count = family.count()

    with pytest.raises(TypeError, match=r'only nodes of types `.*` can be added'):
        family.add_nodes(nodes_incorrect_type)

    assert family.count() == count


@pytest.mark.usefixtures('clear_db')
def test_add_nodes_duplicate_element(get_pseudo_family, get_pseudo_potential_data):
    """Test that `PseudoPotentialFamily.add_nodes` fails if a pseudo is added whose element already exists."""
    family = get_pseudo_family(elements=('Ar',))
    pseudo = get_pseudo_potential_data('Ar').store()

    with pytest.raises(ValueError, match='element `Ar` already present in this family'):
        family.add_nodes(pseudo)


@pytest.mark.usefixtures('clear_db')
def test_remove_nodes(get_pseudo_family):
    """Test the ``PseudoPotentialFamily.remove_nodes`` method."""
    elements = ('Ar', 'He', 'Kr')
    family = get_pseudo_family(elements=elements)
    pseudos = family.get_pseudos(elements=elements)

    # Removing a single node
    pseudo = pseudos.pop('Ar')
    family.remove_nodes(pseudo)
    assert family.pseudos == pseudos

    # Removing multiple nodes
    family.remove_nodes(list(pseudos.values()))
    assert family.pseudos == {}


@pytest.mark.usefixtures('clear_db')
def test_remove_nodes_not_existing(get_pseudo_family, get_pseudo_potential_data):
    """Test the ``PseudoPotentialFamily.remove_nodes`` method works even when passing a non-existing pseudo.

    The implementation of the ``remove_nodes`` method of the ``Group`` base class does not raise when passing a node
    that is not contained within the group but will silently ignore it. Make sure that the corresponding element is not
    accidentally still removed by the ``PseudoPotentialFamily.remove_nodes`` implementation.
    """
    element = 'Ar'
    family = get_pseudo_family(elements=(element,))
    pseudo = get_pseudo_potential_data(element).store()

    # The node ``pseudo`` is not actually contained within the family and so no pseudopotentials should be removed
    family.remove_nodes(pseudo)
    assert tuple(family.pseudos.keys()) == ('Ar',)


@pytest.mark.usefixtures('clear_db')
def test_clear(get_pseudo_family):
    """Test the ``PseudoPotentialFamily.clear`` method."""
    family = get_pseudo_family(elements=('Ar', 'He', 'Kr'))
    assert family.pseudos is not None

    family.clear()
    assert family.pseudos == {}
    assert family.count() == 0


@pytest.mark.usefixtures('clear_db')
def test_pseudos(get_pseudo_potential_data):
    """Test the `PseudoPotentialFamily.pseudos` property."""
    pseudos = {
        'Ar': get_pseudo_potential_data('Ar').store(),
        'He': get_pseudo_potential_data('He').store(),
    }
    family = PseudoPotentialFamily(label='label').store()
    family.add_nodes(list(pseudos.values()))
    assert family.pseudos == pseudos


@pytest.mark.usefixtures('clear_db')
def test_pseudos_mutate(get_pseudo_family, get_pseudo_potential_data):
    """Test that `PseudoPotentialFamily.pseudos` property does not act as a setter."""
    family = get_pseudo_family()

    with pytest.raises(AttributeError):
        family.pseudos = {'He': get_pseudo_potential_data('He')}


@pytest.mark.usefixtures('clear_db')
def test_elements(get_pseudo_family):
    """Test the `PseudoPotentialFamily.elements` property."""
    elements = ['Ar', 'He']
    family = get_pseudo_family(elements=elements)
    assert sorted(family.elements) == elements

    family = PseudoPotentialFamily(label='empty').store()
    assert family.elements == []


@pytest.mark.usefixtures('clear_db')
def test_get_pseudo(get_pseudo_family):
    """Test the `PseudoPotentialFamily.get_pseudo` method."""
    element = 'Ar'
    family = get_pseudo_family(elements=(element,))

    assert family.get_pseudo(element) == family.pseudos[element]

    with pytest.raises(ValueError, match=r'family `.*` does not contain pseudo for element `.*`'):
        family.get_pseudo('He')


@pytest.mark.usefixtures('clear_db')
def test_get_pseudos_raise(get_pseudo_family, generate_structure):
    """Test the `PseudoPotentialFamily.get_pseudos` method when it is supposed to raise."""
    elements = ('Ar', 'He', 'Ne')
    structure = generate_structure(elements)
    family = get_pseudo_family(elements=elements[:2])  # Create family with only subset of the elements

    with pytest.raises(ValueError, match='have to specify one of the keyword arguments `elements` and `structure`.'):
        family.get_pseudos()

    with pytest.raises(ValueError, match='cannot specify both keyword arguments `elements` and `structure`.'):
        family.get_pseudos(elements=elements, structure=structure)

    with pytest.raises(ValueError, match='elements should be a list or tuple of symbols.'):
        family.get_pseudos(elements={'He', 'Ar'})

    with pytest.raises(ValueError, match='structure should be a `StructureData` instance.'):
        family.get_pseudos(structure={'He', 'Ar'})

    with pytest.raises(ValueError, match=r'family `.*` does not contain pseudo for element `.*`'):
        family.get_pseudos(elements=elements)

    with pytest.raises(ValueError, match=r'family `.*` does not contain pseudo for element `.*`'):
        family.get_pseudos(structure=structure)


@pytest.mark.usefixtures('clear_db')
def test_get_pseudos_list(get_pseudo_family):
    """Test the `PseudoPotentialFamily.get_pseudos` method when passing a list of elements."""
    elements = ('Ar', 'He', 'Ne')
    family = get_pseudo_family(elements=elements)

    pseudos = family.get_pseudos(elements=elements)
    assert isinstance(pseudos, dict)
    for element in elements:
        assert isinstance(pseudos[element], PseudoPotentialData)


@pytest.mark.usefixtures('clear_db')
def test_get_pseudos_structure(get_pseudo_family, generate_structure):
    """Test the `PseudoPotentialFamily.get_pseudos` method when passing a ``StructureData`` instance."""
    elements = ('Ar', 'He', 'Ne')
    structure = generate_structure(elements)
    family = get_pseudo_family(elements=elements)

    pseudos = family.get_pseudos(structure=structure)
    assert isinstance(pseudos, dict)
    for element in elements:
        assert isinstance(pseudos[element], PseudoPotentialData)


@pytest.mark.usefixtures('clear_db')
def test_get_pseudos_structure_kinds(get_pseudo_family, generate_structure):
    """Test the `PseudoPotentialFamily.get_pseudos` for ``StructureData`` with kind names including digits."""
    elements = ('Ar1', 'Ar2')
    structure = generate_structure(elements)
    family = get_pseudo_family(elements=elements)

    pseudos = family.get_pseudos(structure=structure)
    assert isinstance(pseudos, dict)
    for element in elements:
        assert isinstance(pseudos[element], PseudoPotentialData)



# -*- coding: utf-8 -*-
# pylint: disable=redefined-outer-name
"""Tests for the :py:`~aiida_pseudo.data.pseudo.jthxml` module."""
import io
import os
import pathlib

import pytest

from aiida.common.exceptions import ModificationNotAllowed
from aiida_pseudo.data.pseudo import JthXmlData


@pytest.fixture
def source(request, filepath_pseudos):
    """Return a pseudopotential, eiter as ``str``, ``Path`` or ``io.BytesIO``."""
    filepath_pseudo = pathlib.Path(filepath_pseudos(entry_point='jthxml')) / 'Ar.jthxml'

    if request.param is str:
        return str(filepath_pseudo)

    if request.param is pathlib.Path:
        return filepath_pseudo

    return io.BytesIO(filepath_pseudo.read_bytes())


@pytest.mark.parametrize('source', (io.BytesIO, str, pathlib.Path), indirect=True)
def test_constructor_source_types(source):
    """Test the constructor accept the various types."""
    pseudo = JthXmlData(source)
    assert isinstance(pseudo, JthXmlData)
    assert not pseudo.is_stored


def test_constructor(filepath_pseudos):
    """Test the constructor."""
    for filename in os.listdir(filepath_pseudos('jthxml')):
        with open(os.path.join(filepath_pseudos('jthxml'), filename), 'rb') as handle:
            pseudo = JthXmlData(handle, filename=filename)
            assert isinstance(pseudo, JthXmlData)
            assert not pseudo.is_stored
            assert pseudo.element == filename.split('.')[0]


@pytest.mark.usefixtures('clear_db')
def test_set_file(filepath_pseudos, get_pseudo_potential_data):
    """Test the `JthXmlData.set_file` method.

    This method allows to change the file, as long as the node has not been stored yet. We need to verify that all the
    information, such as attributes are commensurate when it is stored.
    """
    pseudo = get_pseudo_potential_data(element='Ar', entry_point='jthxml')
    assert pseudo.element == 'Ar'

    with open(os.path.join(filepath_pseudos('jthxml'), 'He.jthxml'), 'rb') as handle:
        pseudo.set_file(handle)
        assert pseudo.element == 'He'

        pseudo.store()
        assert pseudo.element == 'He'

        with pytest.raises(ModificationNotAllowed):
            pseudo.set_file(handle)



# -*- coding: utf-8 -*-
"""Reusable arguments for CLI commands."""
from aiida.cmdline.params.arguments import OverridableArgument
from .types import PseudoPotentialFamilyParam

__all__ = ('PSEUDO_POTENTIAL_FAMILY')

PSEUDO_POTENTIAL_FAMILY = OverridableArgument(
    'family', type=PseudoPotentialFamilyParam(sub_classes=('aiida.groups:pseudo.family',))
)



# -*- coding: utf-8 -*-
"""Reusable options for CLI commands."""
import shutil

import click

from aiida.cmdline.params.options import OverridableOption
from .types import PseudoPotentialFamilyTypeParam, PseudoPotentialTypeParam, UnitParamType

__all__ = (
    'VERSION', 'FUNCTIONAL', 'RELATIVISTIC', 'PROTOCOL', 'PSEUDO_FORMAT', 'STRINGENCY', 'DEFAULT_STRINGENCY',
    'TRACEBACK', 'FAMILY_TYPE', 'ARCHIVE_FORMAT', 'UNIT', 'DOWNLOAD_ONLY'
)

VERSION = OverridableOption(
    '-v', '--version', type=click.STRING, required=False, help='Select the version of the installed configuration.'
)

FUNCTIONAL = OverridableOption(
    '-x',
    '--functional',
    type=click.STRING,
    required=False,
    help='Select the functional of the installed configuration.'
)

RELATIVISTIC = OverridableOption(
    '-r',
    '--relativistic',
    type=click.STRING,
    required=False,
    help='Select the type of relativistic effects included in the installed configuration.'
)

PROTOCOL = OverridableOption(
    '-p', '--protocol', type=click.STRING, required=False, help='Select the protocol of the installed configuration.'
)

PSEUDO_FORMAT = OverridableOption(
    '-f',
    '--pseudo-format',
    type=click.STRING,
    required=True,
    help='Select the pseudopotential file format of the installed configuration.'
)

STRINGENCY = OverridableOption(
    '-s', '--stringency', type=click.STRING, required=False, help='Stringency level for the recommended cutoffs.'
)

DEFAULT_STRINGENCY = OverridableOption(
    '-s',
    '--default-stringency',
    type=click.STRING,
    required=False,
    help='Select the default stringency level for the installed configuration.'
)

TRACEBACK = OverridableOption(
    '-t', '--traceback', is_flag=True, help='Include the stacktrace if an exception is encountered.'
)

FAMILY_TYPE = OverridableOption(
    '-F',
    '--family-type',
    type=PseudoPotentialFamilyTypeParam(),
    default='pseudo.family',
    show_default=True,
    help='Choose the type of pseudo potential family to create.'
)

PSEUDO_TYPE = OverridableOption(
    '-P',
    '--pseudo-type',
    type=PseudoPotentialTypeParam(),
    default='pseudo',
    show_default=True,
    help=(
        'Select the pseudopotential type to be used for the family. Should be the entry point name of a '
        'subclass of `PseudoPotentialData`.'
    )
)

ARCHIVE_FORMAT = OverridableOption(
    '-f', '--archive-format', type=click.Choice([fmt[0] for fmt in shutil.get_archive_formats()])
)

UNIT = OverridableOption(
    '-u',
    '--unit',
    type=UnitParamType(quantity='energy'),
    required=False,
    default='eV',
    show_default=True,
    help='Specify the energy unit of the cutoffs. Must be recognized by the ``UnitRegistry`` of the ``pint`` library.'
)

DOWNLOAD_ONLY = OverridableOption(
    '--download-only',
    is_flag=True,
    help=(
        'Only download the pseudopotential files to the current working directory, without installing the '
        'pseudopotential family.'
    )
)



# -*- coding: utf-8 -*-
# pylint: disable=redefined-outer-name
"""Tests for `aiida-pseudo install`."""
import json
import pathlib

import pytest

from aiida.orm import QueryBuilder

from aiida_pseudo.cli import install
from aiida_pseudo.cli import cmd_install_family, cmd_install_sssp, cmd_install_pseudo_dojo
from aiida_pseudo.data.pseudo.upf import UpfData
from aiida_pseudo.groups.family import PseudoPotentialFamily
from aiida_pseudo.groups.family.pseudo_dojo import PseudoDojoFamily, PseudoDojoConfiguration
from aiida_pseudo.groups.family.sssp import SsspFamily, SsspConfiguration


@pytest.fixture
def run_monkeypatched_install_sssp(run_cli_command, get_pseudo_potential_data, monkeypatch, tmpdir):
    """Fixture to monkeypatch the ``aiida_pseudo.cli.install.download_sssp`` method and call the install cmd."""

    def download_sssp(
        configuration: SsspConfiguration,
        filepath_archive: pathlib.Path,
        filepath_metadata: pathlib.Path,
        traceback: bool = False
    ) -> None:
        # pylint: disable=unused-argument
        """Download the pseudopotential archive and metadata for an SSSP configuration to a path on disk.

        :param configuration: the SSSP configuration to download.
        :param filepath_archive: absolute filepath to write the pseudopotential archive to.
        :param filepath_metadata: absolute filepath to write the metadata file to.
        :param traceback: boolean, if true, print the traceback when an exception occurs.
        """
        import hashlib
        import shutil

        element = 'Ar'
        pseudo = get_pseudo_potential_data(element)
        filepath = tmpdir / pseudo.filename

        with pseudo.open(mode='rb') as handle:
            md5 = hashlib.md5(handle.read()).hexdigest()
            handle.seek(0)
            filepath.write_binary(handle.read())

        filename_archive = shutil.make_archive('temparchive', 'gztar', root_dir=tmpdir, base_dir='.')
        shutil.move(pathlib.Path.cwd() / filename_archive, filepath_archive)

        with open(filepath_metadata, 'w') as handle:
            data = {element: {'md5': md5, 'cutoff_wfc': 60.0, 'cutoff_rho': 240.0}}
            json.dump(data, handle)
            handle.flush()

    def _run_monkeypatched_install_sssp(options=None, raises=None):
        monkeypatch.setattr(install, 'download_sssp', download_sssp)
        return run_cli_command(cmd_install_sssp, options, raises)

    return _run_monkeypatched_install_sssp


@pytest.fixture
def run_monkeypatched_install_pseudo_dojo(run_cli_command, get_pseudo_potential_data, monkeypatch, tmpdir):
    """Fixture to monkeypatch the ``aiida_pseudo.cli.install.download_pseudo_dojo`` method and call the install cmd."""

    def download_pseudo_dojo(
        configuration: PseudoDojoConfiguration,
        filepath_archive: pathlib.Path,
        filepath_metadata: pathlib.Path,
        traceback: bool = False
    ) -> None:
        # pylint: disable=unused-argument
        """Download the pseudopotential archive and metadata for a PseudoDojo configuration to a path on disk.

        :param configuration: the PseudoDojo configuration to download.
        :param filepath_archive: absolute filepath to write the pseudopotential archive to.
        :param filepath_metadata: absolute filepath to write the metadata archive to.
        :param traceback: boolean, if true, print the traceback when an exception occurs.
        """
        import hashlib
        import shutil

        element = 'Ar'
        pseudo = get_pseudo_potential_data(element, entry_point='jthxml')
        filepath = tmpdir / pseudo.filename

        with pseudo.open(mode='rb') as handle:
            md5 = hashlib.md5(handle.read()).hexdigest()
            handle.seek(0)
            filepath.write_binary(handle.read())

        filename_archive = shutil.make_archive('temparchive', 'gztar', root_dir=tmpdir, base_dir='.')
        shutil.move(pathlib.Path.cwd() / filename_archive, filepath_archive)

        data = {'hints': {'high': {'ecut': 20.00}, 'low': {'ecut': 20.00}, 'normal': {'ecut': 20.00}}, 'md5': md5}

        filepath_djrepo = tmpdir / f'{element}.djrepo'

        with open(filepath_djrepo, 'w') as handle:
            json.dump(data, handle)
            handle.flush()

        filename_metadata = shutil.make_archive('tempmetadata', 'gztar', root_dir=tmpdir, base_dir='.')
        shutil.move(pathlib.Path.cwd() / filename_metadata, filepath_metadata)

    def _run_monkeypatched_install_pseudo_dojo(options=None, raises=None):
        monkeypatch.setattr(install, 'download_pseudo_dojo', download_pseudo_dojo)
        return run_cli_command(cmd_install_pseudo_dojo, options, raises)

    return _run_monkeypatched_install_pseudo_dojo


@pytest.mark.usefixtures('clear_db')
def test_install_family(run_cli_command, get_pseudo_archive):
    """Test ``aiida-pseudo install family``."""
    label = 'family'
    description = 'description'
    filepath_archive = next(get_pseudo_archive())
    options = ['-D', description, filepath_archive, label]

    result = run_cli_command(cmd_install_family, options)
    assert f'installed `{label}`' in result.output
    assert PseudoPotentialFamily.objects.count() == 1

    family = PseudoPotentialFamily.objects.get(label=label)
    assert family.__class__ is PseudoPotentialFamily
    assert family.description == description
    assert len(family.pseudos) != 0


@pytest.mark.usefixtures('clear_db')
def test_install_family_folder(run_cli_command, filepath_pseudos):
    """Test ``aiida-pseudo install family` from folder`."""
    label = 'family_test'
    description = 'description'
    dirpath = filepath_pseudos()
    options = ['-D', description, dirpath, label]

    result = run_cli_command(cmd_install_family, options)
    assert f'installed `{label}`' in result.output
    assert PseudoPotentialFamily.objects.count() == 1

    family = PseudoPotentialFamily.objects.get(label=label)
    assert family.__class__ is PseudoPotentialFamily
    assert family.description == description
    assert len(family.pseudos) != 0


@pytest.mark.usefixtures('clear_db')
def test_install_family_url(run_cli_command):
    """Test ``aiida-pseudo install family`` when installing from a URL.

    When a URL is passed, the parameter converts it into a ``http.client.HTTPResponse``, which is not trivial to mock so
    instead we use an actual URL, which is slow, but is anyway already tested indirectly in ``test_install_sssp``.
    """
    label = 'SSSP/1.0/PBE/efficiency'
    description = 'description'
    filepath_archive = 'https://legacy-archive.materialscloud.org/file/2018.0001/v4/SSSP_1.0_PBE_efficiency.tar.gz'
    options = ['-D', description, filepath_archive, label, '-P', 'pseudo.upf']

    result = run_cli_command(cmd_install_family, options)
    assert f'installed `{label}`' in result.output
    assert PseudoPotentialFamily.objects.count() == 1

    family = PseudoPotentialFamily.objects.get(label=label)
    assert isinstance(family.pseudos['Si'], UpfData)
    assert family.__class__ is PseudoPotentialFamily
    assert family.description == description
    assert len(family.pseudos) != 0


@pytest.mark.usefixtures('clear_db')
def test_install_sssp(run_cli_command):
    """Test the ``aiida-pseudo install sssp`` command."""
    from aiida_pseudo import __version__

    result = run_cli_command(cmd_install_sssp)
    assert 'installed `SSSP/' in result.output
    assert QueryBuilder().append(SsspFamily).count() == 1

    family = QueryBuilder().append(SsspFamily).one()[0]
    assert family.get_cutoffs is not None
    assert f'SSSP v1.1 PBE efficiency installed with aiida-pseudo v{__version__}' in family.description
    assert 'Archive pseudos md5: 4803ce9fd1d84c777f87173cd4a2de33' in family.description
    assert 'Pseudo metadata md5: 0d5d6c2c840383c7c4fc3a99b5dc3001' in family.description

    result = run_cli_command(cmd_install_sssp, raises=SystemExit)
    assert 'is already installed' in result.output


@pytest.mark.usefixtures('clear_db')
def test_install_pseudo_dojo(run_cli_command):
    """Test the ``aiida-pseudo install pseudo-dojo`` command."""
    from aiida_pseudo import __version__

    result = run_cli_command(cmd_install_pseudo_dojo)
    assert 'installed `PseudoDojo/' in result.output
    assert QueryBuilder().append(PseudoDojoFamily).count() == 1

    family = QueryBuilder().append(PseudoDojoFamily).one()[0]
    assert family.get_cutoffs is not None
    assert f'PseudoDojo v0.4 PBE SR standard psp8 installed with aiida-pseudo v{__version__}' in family.description
    assert 'Archive pseudos md5: a43737369e8a0a4417ccf364397298b3' in family.description
    assert 'Pseudo metadata md5: d0c0057f16cb905bb2d43382146ffad2' in family.description

    result = run_cli_command(cmd_install_pseudo_dojo, raises=SystemExit)
    assert 'is already installed' in result.output


@pytest.mark.usefixtures('clear_db')
def test_install_sssp_monkeypatched(run_monkeypatched_install_sssp):
    """Test the ``aiida-pseudo install sssp`` command with a monkeypatched download function.

    This circumvents the actual download from Materials Cloud archive and substitutes it with a simple archive of a
    single pseudo potential. We are merely verifying that the various configuration options are respected and are
    reflected in the label of the created family.
    """
    version = '1.1'
    functional = 'PBEsol'
    protocol = 'precision'
    configuration = SsspConfiguration(version, functional, protocol)
    label = SsspFamily.format_configuration_label(configuration)

    options = ['-v', version, '-x', functional, '-p', protocol]
    result = run_monkeypatched_install_sssp(options=options)
    assert f'installed `{label}`' in result.output
    assert SsspFamily.objects.count() == 1

    family = SsspFamily.objects.get(label=label)
    assert family.label == label


@pytest.mark.usefixtures('clear_db')
def test_install_pseudo_dojo_monkeypatched(run_monkeypatched_install_pseudo_dojo):
    """Test the ``aiida-pseudo install pseudo-dojo`` command with a monkeypatched download function.

    This circumvents the actual download from the Pseudo Dojo website and substitutes it with a simple archive of a
    single pseudo potential. We are merely verifying that the various configuration options are respected and are
    reflected in the label of the created family.
    """
    version = '1.0'
    functional = 'LDA'
    protocol = 'stringent'
    pseudo_format = 'jthxml'
    configuration = PseudoDojoConfiguration(version, functional, 'SR', protocol, pseudo_format)
    label = PseudoDojoFamily.format_configuration_label(configuration)

    options = ['-v', version, '-x', functional, '-p', protocol, '-f', pseudo_format]
    result = run_monkeypatched_install_pseudo_dojo(options=options)
    assert f'installed `{label}`' in result.output
    assert PseudoDojoFamily.objects.count() == 1

    family = PseudoDojoFamily.objects.get(label=label)
    assert family.label == label


@pytest.mark.usefixtures('clear_db', 'chtmpdir')
def test_install_sssp_download_only(run_monkeypatched_install_sssp):
    """Test the ``aiida-pseudo install sssp`` command with the ``--download-only`` option."""
    options = ['--download-only']
    result = run_monkeypatched_install_sssp(options=options)

    assert SsspFamily.objects.count() == 0
    assert 'written to the current directory.' in result.output


@pytest.mark.usefixtures('clear_db', 'chtmpdir')
def test_install_sssp_download_only_exists(run_monkeypatched_install_sssp, get_pseudo_family):
    """Test the ``aiida-pseudo install sssp`` command with the ``--download-only`` option.

    The files should be downloadable even if the corresponding configuration is already installed.
    """
    version = '1.1'
    functional = 'PBEsol'
    protocol = 'precision'

    # Generate a family with the same label as we will install through the CLI
    label = SsspFamily.format_configuration_label(SsspConfiguration(version, functional, protocol))
    get_pseudo_family(cls=SsspFamily, pseudo_type=UpfData, label=label)

    options = ['--download-only', '-v', version, '-x', functional, '-p', protocol]
    result = run_monkeypatched_install_sssp(options=options)

    assert SsspFamily.objects.count() == 1
    assert 'written to the current directory.' in result.output


@pytest.mark.usefixtures('clear_db', 'chtmpdir')
def test_install_pseudo_dojo_download_only(run_monkeypatched_install_pseudo_dojo):
    """Test the ``aiida-pseudo install pseudo-dojo`` command with the ``--download-only`` option."""
    options = ['--download-only']
    result = run_monkeypatched_install_pseudo_dojo(options=options)

    assert PseudoDojoFamily.objects.count() == 0
    assert 'written to the current directory.' in result.output


@pytest.mark.usefixtures('clear_db', 'chtmpdir')
def test_install_pseudo_dojo_download_only_exists(run_monkeypatched_install_pseudo_dojo, get_pseudo_family):
    """Test the ``aiida-pseudo install pseudo_dojo`` command with the ``--download-only`` option.

    The files should be downloadable even if the corresponding configuration is already installed.
    """
    version = '1.0'
    functional = 'LDA'
    relativistic = 'SR'
    protocol = 'stringent'
    pseudo_format = 'jthxml'

    # Generate a family with the same label as we will install through the CLI
    configuration = PseudoDojoConfiguration(version, functional, relativistic, protocol, pseudo_format)
    label = PseudoDojoFamily.format_configuration_label(configuration)
    get_pseudo_family(cls=PseudoDojoFamily, pseudo_type=UpfData, label=label)

    options = [
        '--download-only', '-v', version, '-x', functional, '-r', relativistic, '-p', protocol, '-f', pseudo_format
    ]
    result = run_monkeypatched_install_pseudo_dojo(options=options)

    assert PseudoDojoFamily.objects.count() == 1
    assert 'written to the current directory.' in result.output



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument,pointless-statement
"""Tests for the `PseudoDojoFamily` class."""
import pytest

from aiida_pseudo.data.pseudo import UpfData, Psp8Data, PsmlData, JthXmlData
from aiida_pseudo.groups.family import PseudoDojoConfiguration, PseudoDojoFamily


def test_type_string(clear_db):
    """Verify the `_type_string` class attribute is correctly set to the corresponding entry point name."""
    assert PseudoDojoFamily._type_string == 'pseudo.family.pseudo_dojo'  # pylint: disable=protected-access


def test_pseudo_types():
    """Test the `PseudoDojoFamily.pseudo_types` method."""
    assert PseudoDojoFamily.pseudo_types == (UpfData, PsmlData, Psp8Data, JthXmlData)


def test_default_configuration():
    """Test the `PseudoDojoFamily.default_configuration` class attribute."""
    assert isinstance(PseudoDojoFamily.default_configuration, PseudoDojoConfiguration)


def test_valid_configurations():
    """Test the `PseudoDojoFamily.valid_configurations` class attribute."""
    valid_configurations = PseudoDojoFamily.valid_configurations
    assert isinstance(valid_configurations, tuple)

    for entry in valid_configurations:
        assert isinstance(entry, PseudoDojoConfiguration)


def test_get_valid_labels():
    """Test the `PseudoDojoFamily.get_valid_labels` class method."""
    valid_labels = PseudoDojoFamily.get_valid_labels()
    assert isinstance(valid_labels, tuple)

    for entry in valid_labels:
        assert isinstance(entry, str)


def test_format_configuration_label():
    """Test the `PseudoDojoFamily.format_configuration_label` class method."""
    configuration = PseudoDojoConfiguration('0.4', 'PBE', 'SR', 'standard', 'psp8')
    assert PseudoDojoFamily.format_configuration_label(configuration) == 'PseudoDojo/0.4/PBE/SR/standard/psp8'


def test_constructor():
    """Test that the `PseudoDojoFamily` constructor validates the label."""
    with pytest.raises(ValueError, match=r'the label `.*` is not a valid PseudoDojo configuration label'):
        PseudoDojoFamily()

    with pytest.raises(ValueError, match=r'the label `.*` is not a valid PseudoDojo configuration label'):
        PseudoDojoFamily(label='nc-sr-04_pbe_standard_psp8')

    label = PseudoDojoFamily.format_configuration_label(PseudoDojoFamily.default_configuration)
    family = PseudoDojoFamily(label=label)
    assert isinstance(family, PseudoDojoFamily)


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder(filepath_pseudos):
    """Test the `PseudoDojoFamily.create_from_folder` class method."""
    family = PseudoDojoFamily.create_from_folder(
        filepath_pseudos('upf'), 'PseudoDojo/0.4/PBE/SR/standard/psp8', pseudo_type=UpfData
    )
    assert isinstance(family, PseudoDojoFamily)


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_duplicate(filepath_pseudos):
    """Test that `PseudoDojoFamily.create_from_folder` raises for duplicate label."""
    label = 'PseudoDojo/0.4/PBE/SR/standard/psp8'
    PseudoDojoFamily(label=label).store()

    with pytest.raises(ValueError, match=r'the PseudoDojoFamily `.*` already exists'):
        PseudoDojoFamily.create_from_folder(filepath_pseudos('upf'), label)
